
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "common.h"
#include "buspirate.h"
#include "usbasp.h"

#include <usb.h>

#define USB_TIMEOUT     5000

// PIC extensions to USBASP
#define USBAPSP_FUNC_MCLR            17			// set the nMCLR state
#define USBAPSP_FUNC_SENDBITS        18         // send some bits
#define USBAPSP_FUNC_CMD_SEND16      19			// PIC18 write 16 bits
#define USBAPSP_FUNC_CMD_READ16      20         // PIC18 read 8bits
#define USBAPSP_FUNC_SIX             21         // PIC24 send
#define USBAPSP_FUNC_REQOUT          22         // PIC24 read
#define USBAPSP_FUNC_INIT_ICSP       23         // setup hardware for ICSP
#define USBAPSP_FUNC_PGC             24         // set PGC state
#define USBAPSP_FUNC_PGD             25         // set PGD state

#define VENDORID        0x16c0
#define PRODUCTID       0x05dd		// change that to usual USBASP later

#define CMD1(dptr, cmd, data) \
	*((dptr)+0) = ((cmd) & 15) << 4; \
	*((dptr)+1) = (data) & 255; \
	*((dptr)+2) = ((data) >> 8) & 255; \
	dptr += 3

#define CMD2(dptr, cmd, data) \
	*((dptr)-3) |= (cmd) & 15; \
	*((dptr)+0) = (data) & 255; \
	*((dptr)+1) = ((data) >> 8) & 255; \
	dptr += 2

// private
static usb_dev_handle *hdev;
static uint8_t icsp_data[256];
static int     icsp_ptr;

static uint32_t usbasp_Init(struct picprog_t *p, char *port, char *speed)
{
	struct usb_bus *pbus;
	struct usb_device *pdev;
	int ret;
	
	// initialize command pool
	memset(icsp_data, 0, 256);
	icsp_ptr = 0;
	
	// initialize usb device
	usb_init();

	usb_find_busses();
	usb_find_devices();

	hdev = NULL;
	for(pbus = usb_get_busses(); pbus; pbus = pbus->next)
	for(pdev = pbus->devices; pdev; pdev = pdev->next) {
		if ((pdev->descriptor.idVendor == VENDORID) &&
		    (pdev->descriptor.idProduct == PRODUCTID))
		{
			hdev = usb_open(pdev);
		//	break;		// are we too fast? dbl-break?
			goto configured;
		}
	}

configured:
	if (hdev == NULL) {
		printf("device not found\n");
		return 1;
	}

	ret = usb_set_configuration(hdev, 1);
	if (ret < 0) {
		printf("error: could not select configuration (1)\n");
		return -ret;
	}

	ret = usb_claim_interface(hdev, 0);
	if (ret < 0) {
		printf("error: could not claim interface #0\n");
		return -ret;
	}
/*
	// to complete the hack
	ret = usb_control_msg(hdev, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
			USBAPSP_FUNC_INIT_ICSP, 0x0A12, 0xc2b2, NULL, 0, USB_TIMEOUT);
	if (ret < 0) {
		printf("error: ICSP init failed\n");
		return -ret;
	}
*/
	return 0;
}

static uint32_t usbasp_Deinit(struct picprog_t *p)
{
	// close usb connection
	if (hdev != NULL) {
		usb_close(hdev);
		hdev = NULL;
	}
	
//	printf(" <<<<<<<<<<<<< USBASP DISCONNECT >>>>>>>>>>>>>> \n");
	
	return 0;
}

// n/a
static uint32_t usbasp_VPPHigh()
{
	printf("############ N/A ############ - usbasp_VPPHigh\n");
	return 0;
}

// n/a
static uint32_t usbasp_VPPLow()
{
	printf("############ N/A ############ - usbasp_VPPLow\n");
	return 0;
}

static uint32_t usbasp_DataLow()
{
	// swing low sweet chariot
	if (hdev == NULL) return -1;
	return usb_control_msg(hdev, 0x40, USBAPSP_FUNC_PGD, 0, 0, NULL, 0, USB_TIMEOUT);
}

static uint32_t usbasp_DataHigh()
{
	if (hdev == NULL) return -1;
	return usb_control_msg(hdev, 0x40, USBAPSP_FUNC_PGD, 0xffff, 0, NULL, 0, USB_TIMEOUT);
}

static uint32_t usbasp_ClockLow()
{
	if (hdev == NULL) return -1;
	return usb_control_msg(hdev, 0x40, USBAPSP_FUNC_PGC, 0, 0, NULL, 0, USB_TIMEOUT);
}

static uint32_t usbasp_ClockHigh()
{
	if (hdev == NULL) return -1;
	return usb_control_msg(hdev, 0x40, USBAPSP_FUNC_PGC, 0xffff, 0, NULL, 0, USB_TIMEOUT);
}

static uint32_t usbasp_MCLRLow()
{
	if (hdev == NULL) return -1;
	return usb_control_msg(hdev, 0x40, USBAPSP_FUNC_MCLR, 0, 0, NULL, 0, USB_TIMEOUT);
}

static uint32_t usbasp_MCLRHigh()
{
	if (hdev == NULL) return -1;
	return usb_control_msg(hdev, 0x40, USBAPSP_FUNC_MCLR, 0xffff, 0, NULL, 0, USB_TIMEOUT);
}

static uint8_t usbasp_reversebyte(uint8_t c)
{
	printf("############ N/A ############ - usbasp_reversebyte\n");
	return 0;
}

static int usbasp_SetPicMode(enum BP_picmode_t mode)
{
	printf("pic18 only\n");
	return 0;
}

static uint32_t usbasp_PIC416Write(uint8_t cmd, uint16_t data)
{
	uint8_t *ptr;
	
	if (cmd & 0xF0) {
		cmd = 3;	// a "special nop"
		data = 0x0101;
	}
	
	ptr = &icsp_data[icsp_ptr];
	switch (icsp_ptr & 7) {
		case 0:
		case 5:
			CMD1(ptr, cmd, data);
			break;
			
		case 3:
			CMD2(ptr, cmd, data);
			break;
	}
	icsp_ptr += ptr - &icsp_data[icsp_ptr];
	
	return 0;
}

static void BPdr(char *Data, uint32_t length)
{
}

static uint32_t usbasp_PIC416Read(uint8_t cmd, void *Data, uint32_t length)
{
	int ret = 0;
	
	ret = usb_control_msg(hdev, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
			USBAPSP_FUNC_CMD_READ16, cmd, 0, Data, length, USB_TIMEOUT);
	
//	dumphex
	
	return 0;
}

static uint32_t usbasp_PIC424Read(uint32_t cmd, void *Data, uint32_t length)
{
	printf("############ N/A ############ - usbasp_PIC424Read\n");
	return 0; //usbasp_reversebyte(buffer[0]) | usbasp_reversebyte(buffer[1]) << 8;	//upper 8 bits
}

static uint32_t usbasp_PIC424Write(uint32_t data, uint8_t prenop, uint8_t postnop)
{
	printf("############ N/A ############ - usbasp_PIC424Write\n");
	return 0;
}

static uint32_t usbasp_Flush()
{
	int ret;
	
//	printf("flush\n");
	ret = usb_control_msg(hdev, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
			USBAPSP_FUNC_CMD_SEND16, 0, 0, &icsp_data[0], icsp_ptr, USB_TIMEOUT);
	icsp_ptr = 0;
	
	return 0;
}

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// ---------------------------------------------------------------------

static uint32_t usbasp_SetBitOrder(uint8_t lsb)
{
	// yo man!
	printf("############ N/A ############ - usbasp_SetBitOrder\n");
	
	return 0;
}

// not implemented (also no used in ppp)
static uint32_t usbasp_WriteToPirate(int fd, char * val)
{
	printf("############ N/A ############ - usbasp_WriteToPirate\n");
	return 0;
}

// n/a
static void usbasp_EnableRaw2Wire(int fd)
{
	printf("############ N/A ############ - usbasp_EnableRaw2Wire\n");
}

void dumphex(uint8_t *buff, int length)
{
        int i;
        
        for (i=0; i<length; i++) {
                if ((i % 16) == 0) {
                        printf("\n%08X : ", i);
                }
                printf("%02X ", buff[i]);
        }
        printf("\n");
}

//binmode: bulk write bytes to bus command
static uint32_t usbasp_BulkByteWrite(uint8_t bwrite, char* val)
{
	int ret;
	// bulky
//	printf("############ N/A ############ - usbasp_BulkByteWrite\n");

//	printf("write %d bytes:", bwrite);
//	dumphex(val, bwrite);
	ret = usb_control_msg(hdev, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
			USBAPSP_FUNC_SENDBITS, bwrite*8, 0, val, bwrite, USB_TIMEOUT);
	
	return 0;
}

static uint32_t usbasp_BulkBitWrite(uint8_t bit_count, char val)
{
	// a bit a time
	printf("############ N/A ############ - usbasp_BulkBitWrite\n");
	
	return 0;
}

struct iface_t usbasp_iface = 
{
	.name = "USBASP",
	.caps = 0,
	.Init = usbasp_Init,
	.Deinit = usbasp_Deinit,
	.DataLow = usbasp_DataLow,
	.DataHigh = usbasp_DataHigh,
	.ClockLow = usbasp_ClockLow,
	.ClockHigh = usbasp_ClockHigh,
	.MCLRLow = usbasp_MCLRLow,
	.MCLRHigh = usbasp_MCLRHigh,
	.SetBitOrder = usbasp_SetBitOrder,
	.SendBytes = usbasp_BulkByteWrite,
	.SendBits = usbasp_BulkBitWrite,
	.PIC416Read = usbasp_PIC416Read,
	.PIC416Write = usbasp_PIC416Write,
	.PIC424Read = usbasp_PIC424Read,
	.PIC424Write = usbasp_PIC424Write,
	.flush=usbasp_Flush,
	.VPPHigh = usbasp_VPPHigh,
	.VPPLow = usbasp_VPPLow,
};
